#!/usr/bin/perl -w
# $Id: backup2disk.pl,v 1.36 2004/09/01 14:56:09 ahi Exp $

# checking lockfile;
$lockfile="/var/lock/backup2tape";
if (-f $lockfile) {
  print "##\n";
  print "## Lockfile '$lockfile' exists\n";
  print "## possibly tape backup in progress\n";
  print "##\n";
  print "## ABORTING\n";
  print "##\n";
  exit -1;
}

# Mountpoint for samba-shares
my $smbmnt = "/mnt/smb";

# Current date
my $date = `date +"%Y-%m-%d"`;
chomp $date;

# PATH to configuration files 
if ($ARGV[0]) {
    $confdir = $ARGV[0];
} else {
    $confdir = "/etc/backup2disk";
}

# Configuration files
my $backup_list    = "$confdir/volumes";
my $backup_samb    = "$confdir/samba";
my $backup_afs     = "$confdir/afs";
my $backup_exclude = "$confdir/excludes";
my $rsync = "rsync -rtgoDuxvzS --exclude-from=$backup_exclude"
  . " --delete --backup"
  . " --suffix=_DELETED_$date --exclude='*_DELETED_????-??-??'";


my %backup_volumes;


use File::Find;
use Sys::Hostname;
use HTTP::Date;
my $hostname = hostname();

($i,$i,$i,$mday,$mon,$i,$i,$i,$i) = localtime(time);
$day = $mday % 7;
$week = int($mday / 7);
$month = $mon % 2;

$|=1;
# --------------------------------------------------------
my ($node, $directory, $maxage, $destination, $time);

print "BACKUP via SSH/RSYNC\n";
umask 0222;
open LIST, "<$backup_list" or die "Unable to open $backup_list: $! ... exiting";
while(<LIST>){
  /^#/ and next;
  /^\s*$/ and next;

  $destination = 0;
  ($node, $directory, $maxage, $destination) = split;
  $destination or warn "incomplete line: $_" and next;
  chomp $destination;
  $backup_volumes{$destination} = 1;     # add volume to list (hash)
  $dest_dir = $directory;
  $dest_dir =~ s/\//-/g;
  $destination .= "/$node$dest_dir";
  $command = ($hostname =~ /^$node/)
    ? "$rsync $directory/ $destination"
    : "$rsync -e ssh $node:$directory/ $destination";
  printf "%9s:%-25s", $node, $directory;

  $time = time;
  $ret = system("$command >/var/log/backup2disk/$node$dest_dir 2>&1");
  $ret = $ret >> 8;
  printf " %7d sec.", time - $time;
  if (0 != $ret) {
    printf " ERROR = %d\n", $ret;
  }
  else {
    finddepth({wanted => \&files }, $destination);
    finddepth({wanted => \&dirs }, $destination);
    printf "\n";
  }
}

print "\nBACKUP via SMB/RSYNC\n";
umask 0277;
-d $smbmnt or mkdir $smbmnt,0755;
system ("umount $smbmnt >/dev/null 2>&1");
open LIST, "<$backup_samb" or die "Unable to open $backup_samb: $! ... exiting";
while(<LIST>){
  /^#/ and next;
  /^\s*$/ and next;

  $passwd = 0;
  ($node, $domain, $directory, $maxage, $destination, $user, $passwd) = split;
  $backup_volumes{$destination} = 1;     # add volume to list (hash)
  $passwd or warn "incomplete line: $_" and next;
  chomp $passwd;

  $ENV{USER}=$user."%".$passwd;
  $dest_dir = $directory;
  $dest_dir =~ s/\//-/g;
  $destination .= "/$node-$dest_dir";
  $command = "$rsync $smbmnt/ $destination";
  printf "%9s:%-25s", $node, $directory;

  $time = time;
  system ("mount -t smbfs -o ro,workgroup=$domain //$node/$directory $smbmnt "
	  .">/var/log/backup2disk/$node-$dest_dir 2>&1");
  `mount | grep $smbmnt`;  # did the mount succed ?????????
  if (0 != $?) {
      print "can't mount: $?\n";
      next;
  }
  $ret = system  ("$command >>/var/log/backup2disk/$node-$dest_dir 2>&1");
  $ret = $ret >> 8;
  system ("umount $smbmnt >>/var/log/backup2disk/$node-$dest_dir 2>&1");
  printf " %7d sec.", time - $time;
  if (0 != $ret) {
    printf " ERROR = %d\n", $ret;
  }
  else {
    finddepth({wanted => \&files }, $destination);
    finddepth({wanted => \&dirs }, $destination);
    printf "\n";
  }
}

print "\nBACKUP via AFS\n";
open LIST, "<$backup_afs" or warn "Unable to open $backup_afs: $! ... continuing";

while (<LIST>){
  /^#/ and next;
  /^\s*$/ and next;
  
  $offset = "";
  ($volset, $volprefix, $dumplevel, $offset) = split;
  
  $offset eq "" and warn "incomplete line: $_" and next;

  printf "%9s:\tportoffset %2d", $volset, $offset;
  $time = time;
  $backupsysed{$volprefix} or
    system  "/usr/sbin/vos backupsys -prefix $volprefix -localauth >& /var/log/backup2disk/vos_$volset";
  $ret = $? >> 8;
  if($ret){
    printf " %7d sec.", time - $time;
    printf " ERROR = vos %d\n", $ret;
    next
  }
  $backupsysed{$volprefix} = 1;

  if($mday == 1){ $dump = "/month$month" }
  elsif($day == 0 && $dumplevel > 0){ $dump = "/month$month/week$week" }
  elsif($dumplevel > 1){ $dump = "/month$month/week$week/day$day"}
  else{
    printf " %7d sec.\n", time - $time;
    next;
  }

  # check for the butc
  open PSLIST, "ps -eo command |" or die "Unable to execute ps -eo command: $!";
  while (<PSLIST>){ /butc $offset/ and $runs = 1 }
  close PSLIST;
  if(! $runs){
    $pid = fork();
    if($pid){push @butcpids, $pid }
    else{
      chdir '/';
      open STDOUT, ">/dev/null" or die "Unable to reopen STDOUT: $!";
      exec "/usr/sbin/butc $offset -localauth >& /var/log/backup2disk/butc.$offset.log";
    }
  }

  # start backuprocess
  system "/usr/sbin/backup dump $volset $dump $offset -localauth >& /var/log/backup2disk/backup_$volset";
  $ret = $? >> 8;
  printf " %7d sec.", time - $time;
  if($ret){
    printf " ERROR = backup %d\n", $ret;
  }else{
    # hardcoded path: /backup ugly !TO_FIX!
    open INFO, "/usr/sbin/backup dumpinfo -ndumps 1000 -localauth 2>/dev/null|"
      or warn "Unable to execute dumpinfo: $!";
    %dumpinfo = ();
    while(<INFO>){
      chomp;
      /(\d+)\s+(\d+)\s+(\d+)\s+([\d\/]+)\s+([\d:]+)\s+(\d+)\s+(\d+)\s+(\S+)/ and $dumpinfo{"$1"} = $1;
    }
    close INFO;
    $destination = "/backup/$offset";
    @compress = ();
    # 
  FILE: foreach (glob "$destination/$volset*"){
      -f or return;
      $fn = $_;
      /$destination\/$volset\.\w+\.\d+_(\d+)/ and $id = $1;
      $id or next;
      if($dumpinfo{$id}){
	open INFO, "/usr/sbin/backup dumpinfo -id $id -verbose -localauth 2>/dev/null|" 
	  or warn "Unable to execute dumpinfo -id $id: $!";
	while(<INFO>){
	  if (/expires = (.*)/) {
	    if (scalar(str2time($1)) < scalar(time())) {
	      `/usr/sbin/backup deletedump $id >/dev/null 2>/dev/null`;
	      delete $dumpinfo{$id};
	      last;
	    }else{
	      if( not ( $fn =~ /\.bz2$/ or $fn =~ /\.gz$/ )){
		push @compress, $fn;
	      }
	      next FILE;
	    }
	  }
	}
      }
      unlink $fn;
    }
    $pid = fork();
    if(!$pid){
      chdir '/';
      open STDOUT, ">/dev/null" or die "Unable to reopen STDOUT: $!";
      exec "/bin/gzip",  "-q", @compress;
    }
    $backup_volumes{$destination} = 1;
    print "\n";
  }
}

@butcpids and kill 15, @butcpids; # Try to kill the forked butc process
`killall butc`;  ################################ HACK ################################


print "\nDestinations for Backup:\n";
@desti = (sort keys %backup_volumes);
print `df -m @desti`;

# sub afsdumps {
#   local $id;
#   local $fn = $_;
#   chomp;
#   -f or return;
#   /^$volset/ or return; # Only delete volumes that have been backed up
#   /($volset\.\w+)\.\d+_(\d+)/ and $id = $2;
#   $id or return;
#   if($dumpinfo{$id}){
#     open INFO, "/usr/sbin/backup dumpinfo -id $id -verbose -localauth 2>/dev/null|" or warn "Unable to execute dumpinfo -id $id: $!";
#     while(<INFO>){
#       if (/expires = (.*)/) {
# 	if (scalar(str2time($1)) < scalar(time())) {
# 	  # print "\n$id has expirytime $1, localtime ", time2str(), " -> deleting!!!\n";
# 	  `/usr/sbin/backup deletedump $id >/dev/null 2>/dev/null`;
# 	  delete $dumpinfo{$id};
# 	  last;
# 	}else{
# 	  if( not ( $fn =~ /\.bz2$/ or $fn =~ /\.gz$/ )){ push @compress, $fn }
# 	  return;
# 	}
#       }
#     }
#   }
#   unlink $fn;
# }

sub files {
  my ($dev,$ino,$mode,$nlink,$uid,$gid);

  /^.*_DELETED_\d{4}-\d{2}-\d{2}$/ &&
    (($dev,$ino,$mode,$nlink,$uid,$gid) = lstat($_)) &&
      (int(-C _) > $maxage ) &&
	-f _ &&
	  unlink;
}

sub dirs {
  my ($dev,$ino,$mode,$nlink,$uid,$gid);

  /^.*_DELETED_\d{4}-\d{2}-\d{2}$/ &&
    (($dev,$ino,$mode,$nlink,$uid,$gid) = lstat($_)) &&
      (int(-C _) > $maxage ) &&
	-d _ &&
	  rmdir;
}


=head1 NAME

backup2disk - create backup copies on a spare disk

=head1 SYNOPSYS

backup2disk [F<config-dir>]

=head1 DESCRIPTION

In case you don't provide a configuration directory on the command line, a
default of F</etc/backup2disk/> is used.

=head2 Backing up filesystems on UNIX hosts (including local filesystems)

The backup ist made by B<rsync>, directories to be backed up are configured
in F</etc/backup2disk/volumes>. The backup server must be able to log into
the remote computer as user C<root> via B<ssh> without user intervention.

=head2 Backing up SMB shares(MS DOS to MS Windows 2000)

smbmount, rsync

=head2 Backing up AFS volume sets

Triggers the AFS backup system - you have to configure AFS backup tape
coordinator to perform backups to disk first. Starting AFS dump is sheduled
by F</etc/backup2disk/afs>.

=head2 Backing up Apple MacIntosh

Unluckily we did not find a practical solution for Apple MacIntosh other than
exporting the UNIX-Homedirectories with netatalk and educating users to place
their data there. Hopefully this situation improves with Mac OS X; installing
OpenSSH and rsync schould not be too difficult then.


=head1 SECURITY

Backing up to a networked disk is no substitution for tape backups.

B<backup2disk> relies heavily on passwordless logins via B<ssh>. Make sure,
your backup server is at least as secure as your file servers.

The file F<samba> holds passwords necessary to mount SMB volumes via samba.
Permissions on this file should be 0700 und it should not reside on an NFS
exported directory.


=head1 FILES

Configuration files are searched in F</etc/backup2disk/> or in a directory
given as an optional parameter to B<backup2disk>. There should be an example
for each of the configuration files installed.

=over

=item F<excludes>

The file F<excludes> lists patterns of filenames not to backup. The format is
as described in manual page rsync(1) under option C<--exclude-from=FILE>.

=item F<volumes>

Every line in F<volumes> describes on combination of UNIX host and directory
to save.

=over

=item 1. name of remote node

If this entry matches the current hostname as obtained by the command
B<hostname>, B<ssh> will not be used - otherwise B<rsync> will open a shh
connection to the named computer.

=item 2. directory

=item 3. maximum age

Describes how long modified or deleted files should be kept.

=item 4. destination of backup

=back

Entries in each line are separated by whitespace, empty
lines and lines starting with "#" are to be ignored.

=item F<samba>

First four entrys have the same meaning as in F<volumes>; the fifth entry
gives a username with enough access rights to read all data und the sixth
entry provides this users password.

=item F<afs>

Every line in F<afs> describes on volumeset with corresponding dumplevel and
destination disk. Entries in each line are separated by whitespace, empty
lines and lines starting with "#" are to be ignored.

=over

=item volumeset

description

=item volumeprefix

description

=item dumplevel

description

=item portoffset

description

=back

=item F</var/lock/backup2tape>

A process writing the backup directories to magnetic tape should set the
lockfile F</var/lock/backup2tape>, to stop B<backup2disk> from changing data.
If the lockfile exists, the B<backup2disk> terminates with return value -1.

=back

=head1 RESTRICTIONS

B<backup2disk> does not backup symbolic links (never checked on hard
links) since strange error messages occur when some links destination
is not known on the backup server.

=head1 BUGS

Script should generate F</var/log/backup2disk/> if it doesnt already
exist. Generates mountpoint F</mnt/smb/> with rather odd permissions.

Unsufficient error checking on B<rsync> return codes. What is retcode 20?

=head1 AUTHORS

=over

=item *

Christian Pfaffel <flash@itp.tugraz.at>

=item *

Andreas Hirczy <ahi@itp.tugraz.at>

=back

=head2 Licensing

Copyright (C) 2001, 2004

B<backup2disk> is distributed under the GNU public license.  See the file
F<COPYING> for details. B<backup2disk> is available from
I<http://itp.tugraz.at/~ahi/Computer/SW/>

=head2 Disclaimer

The authors accept no responsibility for damages resulting from the
use of this software and make no warranty, either expressed or
implied, including but not limited to, any implied warranty of
merchantability or fitness for a particular purpose. This software is
provided as is, and you assume all risks when using it.

=head1 SEE ALSO

rsync(1), ssh(1), sshd(8), smbmount(8)

I<http://www.snailbook.com/faq/no-passphrase.auto.html> for a discussion on
passwordless B<ssh>.

I<http://www.openafs.org/pages/doc/AdminGuide/auagd011.htm> for documentation
on configuring the AFS backup process.

=cut
